from typing import Union

from guniflask.config import settings
from guniflask.security import JwtManager, SecurityContext, BearerTokenExtractor
from guniflask.security_config import SecurityConfigurer, HttpSecurityBuilder
from guniflask.web import RequestFilter
from werkzeug.local import LocalProxy

jwt_manager: Union[JwtManager, LocalProxy] = LocalProxy(lambda: settings['_jwt_manager'])


class JwtConfigurer(SecurityConfigurer):
    def __init__(self):
        super().__init__()
        jwt = settings.get_by_prefix('guniflask.jwt')
        if jwt:
            self.active = True
            settings['_jwt_manager'] = JwtManager(**jwt)
        else:
            self.active = False

    def configure(self, http: HttpSecurityBuilder):
        if self.active:
            http.add_request_filter(JwtFilter())


class JwtFilter(RequestFilter):
    def __init__(self):
        self.token_extractor = BearerTokenExtractor()

    def before_request(self):
        auth = self.token_extractor.extract()
        if auth is not None:
            user_auth = jwt_manager.authenticate(auth)
            SecurityContext.set_authentication(user_auth)



import logging

from flask import Flask
from guniflask.config import Settings
from guniflask.context import AnnotationConfigBeanContext

log = logging.getLogger(__name__)


def make_settings(app: Flask, settings: Settings):
    pass


def init_app(app: Flask, settings: Settings):
    test_bean_context()


def test_bean_context():
    bean_context = AnnotationConfigBeanContext()
    bean_context.scan('rest_app.lifecycle_component')
    bean_context.refresh()
    bean_context.close()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import re
from os.path import join, dirname

# -- Project information -----------------------------------------------------

project = 'guniflask'
copyright = '2018-2021, jadbin'
author = 'jadbin'


# The short X.Y version
def read_version():
    p = join(dirname(dirname(__file__)), 'guniflask', '__init__.py')
    with open(p, 'r', encoding='utf-8') as f:
        return re.search(r"__version__ = '([^']+)'", f.read()).group(1)


version = read_version()
release = version

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

html_theme_options = {
    'description': 'flask + gunicorn, scaffolding tool for web services',
    'github_user': 'jadbin',
    'github_repo': 'guniflask',
    'github_button': False,
    'travis_button': True,
    'font_family': '"Helvetica Neue", Helvetica, "PingFang SC", "Hiragino Sans GB", "Microsoft YaHei", "微软雅黑", Arial, sans-serif',
    'font_size': '14px',
    'code_font_size': '12px',
    'note_bg': '#E5ECD1',
    'note_border': '#BFCF8C',
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_sidebars = {
    '**': [
        'about.html', 'navigation.html', 'searchbox.html',
    ]
}



from flask import abort

from guniflask.config.app_settings import settings
from guniflask.context import configuration, bean
from guniflask.web.bind_annotation import blueprint, get_route


@blueprint
class HealthCheckEndpoint:

    @get_route('/health')
    def health_check(self, app_name: str = None):
        if app_name is not None and settings['app_name'] != app_name:
            abort(400, 'APP name not matched')
        return {'status': 'UP'}


@configuration
class HealthCheckConfiguration:
    @bean
    def heath_check_endpoint(self) -> HealthCheckEndpoint:
        endpoint = HealthCheckEndpoint()
        return endpoint



from abc import ABCMeta, abstractmethod


class UserDetails(metaclass=ABCMeta):

    @property
    @abstractmethod
    def username(self):
        pass  # pragma: no cover

    @property
    @abstractmethod
    def password(self):
        pass  # pragma: no cover

    @property
    @abstractmethod
    def authorities(self):
        pass  # pragma: no cover



import datetime as dt
from abc import ABCMeta, abstractmethod

from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger

from guniflask.beans.lifecycle import SmartInitializingSingleton, DisposableBean


class TaskScheduler(metaclass=ABCMeta):
    @abstractmethod
    def schedule(self, task, start_time: dt.datetime = None):
        pass  # pragma: no cover

    @abstractmethod
    def schedule_with_cron(self, task, cron: str, start_time: dt.datetime = None):
        pass  # pragma: no cover

    @abstractmethod
    def schedule_with_fixed_interval(self, task, interval: int, start_time: dt.datetime = None):
        pass  # pragma: no cover


class DefaultTaskScheduler(TaskScheduler, SmartInitializingSingleton, DisposableBean):
    def __init__(self, scheduler=None):
        if scheduler is None:
            try:
                from apscheduler.schedulers.gevent import GeventScheduler
                self._scheduler = GeventScheduler()
            except ImportError:
                self._scheduler = BackgroundScheduler()
        else:
            self._scheduler = scheduler

    @property
    def scheduler(self):
        return self._scheduler

    def schedule(self, task, start_time: dt.datetime = None):
        trigger = None
        if start_time is not None:
            trigger = DateTrigger(start_time)
        self._scheduler.add_job(task, trigger=trigger)

    def schedule_with_cron(self, task, cron: str, start_time: dt.datetime = None):
        values = cron.split()
        if len(values) != 5:
            raise ValueError(f'Wrong number of fields: got {len(values)}, expected 5')
        cron_trigger = CronTrigger(minute=values[0], hour=values[1], day=values[2], month=values[3],
                                   day_of_week=values[4], start_date=start_time)
        self._scheduler.add_job(task, trigger=cron_trigger)

    def schedule_with_fixed_interval(self, task, interval: int, start_time: dt.datetime = None):
        interval_trigger = IntervalTrigger(seconds=interval, start_date=start_time)
        self._scheduler.add_job(task, trigger=interval_trigger)

    def start(self):
        self._scheduler.start(paused=False)

    def shutdown(self, wait=True):
        self._scheduler.shutdown(wait=wait)

    def after_singletons_instantiated(self):
        self.start()

    def destroy(self):
        self.shutdown(wait=False)



from guniflask.beans import InitializingBean, SmartInitializingSingleton, DisposableBean
from guniflask.context import component


@component
class LifecycleComponent(InitializingBean, SmartInitializingSingleton, DisposableBean):

    def after_properties_set(self):
        pass

    def after_singletons_instantiated(self):
        pass

    def destroy(self):
        pass



from abc import ABCMeta, abstractmethod


class PasswordEncoder(metaclass=ABCMeta):
    @abstractmethod
    def encode(self, raw_password: bytes) -> str:
        pass  # pragma: no cover

    @abstractmethod
    def matches(self, raw_password: bytes, encoded_password: str) -> bool:
        pass  # pragma: no cover


class PlainPasswordEncoder(PasswordEncoder):
    def encode(self, raw_password: bytes) -> str:
        if not raw_password:
            return ''
        return raw_password.decode('utf-8')

    def matches(self, raw_password: bytes, encoded_password: str) -> bool:
        if raw_password is None:
            raw_password = b''
        if encoded_password is None:
            encoded_password = ''
        return raw_password.decode('utf-8') == encoded_password



from guniflask.web import blueprint, get_route


@blueprint('/api')
class MathController:
    def __init__(self, math_service):
        self.math_service = math_service

    @get_route('/math/add')
    def add(self, x: float, y: float):
        return {'result': self.math_service.add(x, y)}



from guniflask.security_config.http_security_builder import HttpSecurityBuilder
from guniflask.security_config.security_configurer import SecurityConfigurer
from guniflask.web.cors import CorsFilter


class CorsConfigurer(SecurityConfigurer):

    def __init__(self, options=None):
        super().__init__()

        if isinstance(options, dict):
            self.cors_filter = CorsFilter(**options)
        else:
            self.cors_filter = CorsFilter()

    def configure(self, http: HttpSecurityBuilder):
        http.add_request_filter(self.cors_filter)



from guniflask.security.user_details import UserDetails


class User(UserDetails):
    def __init__(self, username=None, password=None, authorities=None):
        self._username = username
        self._password = password
        self._authorities = set()
        if authorities:
            self._authorities.update(authorities)

    @property
    def username(self):
        return self._username

    @property
    def password(self):
        return self._password

    @property
    def authorities(self):
        return self._authorities



import builtins
import datetime as dt
import inspect
from collections import OrderedDict
from typing import Any, get_type_hints
from typing import List, Set, Mapping, Optional, Union

from pydantic import BaseModel
from pydantic.typing import get_origin, get_args

from guniflask.utils.datatime import convert_to_datetime


def parse_json(source: Any, dtype: Any = None) -> Any:
    if dtype is None:
        return source
    if source is None:
        return

    target = None
    arg_ = ArgType(dtype)

    if (arg_.is_list() or arg_.is_set()) and not isinstance(source, List):
        source = [source]

    if isinstance(source, Mapping):
        if arg_.is_singleton():
            if issubclass(arg_.outer_type, BaseModel):
                return arg_.outer_type(**source)

            target = arg_.outer_type()
            type_hints = get_type_hints(arg_.outer_type)
            for k, v in source.items():
                v_t = type_hints.get(k)
                vv = parse_json(v, dtype=v_t)
                setattr(target, k, vv)
        elif arg_.is_dict():
            target = dict()
            if arg_.outer_type is not None:
                k_t, v_t = arg_.outer_type
            else:
                k_t = v_t = None
            for k, v in source.items():
                kk = parse_json(k, dtype=k_t)
                vv = parse_json(v, dtype=v_t)
                target[kk] = vv

        if target is None:
            raise AssertionError(f'Cannot convert a dict to {dtype}')

    elif isinstance(source, List):
        collection_type = None
        if arg_.is_list():
            collection_type = list
        elif arg_.is_set():
            collection_type = set

        if collection_type is None:
            raise AssertionError(f'Cannot convert a list to {dtype}')

        target = []
        for v in source:
            target.append(parse_json(v, dtype=arg_.outer_type))
        target = collection_type(target)

    else:
        if arg_.is_singleton():
            if arg_.outer_type == dt.datetime:
                target = convert_to_datetime(source)
            else:
                if type(source) == arg_.outer_type:
                    target = source
                else:
                    target = arg_.outer_type(source)

    return target


def inspect_args(func):
    signature = inspect.signature(func)
    parameters = signature.parameters
    return_type = signature.return_annotation

    hints = {}
    if return_type is not inspect._empty:
        hints['return'] = return_type

    args = OrderedDict()
    for p in parameters.values():
        args[p.name] = p.default
        if p.annotation is not inspect._empty:
            hints[p.name] = p.annotation

    return args, hints


class ArgTypeShape:
    SINGLETON = object()
    DICT = object()
    LIST = object()
    SET = object()
    UNION = object()


class ArgType:
    def __init__(self, arg_type):
        self.arg_type = arg_type

        self.shape = ArgTypeShape.SINGLETON
        self.outer_type = arg_type

        self._analyze_arg_type(arg_type)

    def is_singleton(self):
        return self.shape is ArgTypeShape.SINGLETON

    def is_dict(self):
        return self.shape is ArgTypeShape.DICT

    def is_list(self):
        return self.shape is ArgTypeShape.LIST

    def is_set(self):
        return self.shape is ArgTypeShape.SET

    def is_union(self):
        return self.shape is ArgTypeShape.UNION

    def _analyze_arg_type(self, arg_type: Optional[type]):
        if arg_type is None:
            return

        origin = get_origin(arg_type)
        if origin:
            shape = None
            dtype = None

            if origin is Union:
                allow_types = []

                for arg in get_args(arg_type):
                    if arg is not None:
                        allow_types.append(arg)

                self.shape = ArgTypeShape.UNION
                self.outer_type = [self._get_arg_type(t) for t in allow_types]
                return

            if inspect.isclass(origin):
                if issubclass(origin, List):
                    shape = ArgTypeShape.LIST
                    args = get_args(arg_type)
                    if args and len(args) == 1:
                        vt = args[0]
                        if not inspect.isclass(vt):
                            vt = None
                        dtype = self._get_arg_type(vt)
                elif issubclass(origin, Mapping):
                    shape = ArgTypeShape.DICT
                    args = get_args(arg_type)
                    if args and len(args) == 2:
                        kt = args[0]
                        vt = args[1]
                        dtype = (self._get_arg_type(kt), self._get_arg_type(vt))
                    else:
                        dtype = (None, None)
                elif issubclass(origin, Set):
                    shape = ArgTypeShape.SET
                    args = get_args(arg_type)
                    if args and len(args) == 1:
                        vt = args[0]
                        if not inspect.isclass(vt):
                            vt = None
                        dtype = self._get_arg_type(vt)
            if shape is None:
                raise AssertionError(f'Unsupported generic argument type: {arg_type}')
            self.shape = shape
            self.outer_type = dtype
            return

        if isinstance(arg_type, str):
            if hasattr(builtins, arg_type):
                arg_type = getattr(builtins, arg_type)
                self.outer_type = arg_type

        if not inspect.isclass(arg_type):
            raise AssertionError(f'Non-generic argument type must be a class, but got: {arg_type}')
        if issubclass(arg_type, List):
            self.shape = ArgTypeShape.LIST
            self.outer_type = None
        elif issubclass(arg_type, Mapping):
            self.shape = ArgTypeShape.DICT
            self.outer_type = (None, None)
        elif issubclass(arg_type, Set):
            self.shape = ArgTypeShape.SET
            self.outer_type = None

    def _get_arg_type(self, arg_type):
        try:
            result = self.__class__(arg_type)
        except Exception:
            result = self.__class__(None)
        if result.is_singleton():
            return result.outer_type
        return result


def analyze_arg_type(arg_type: Optional[type]) -> ArgType:
    return ArgType(arg_type)



import gzip
from typing import List

from pydantic import BaseModel

from guniflask.web.request_filter import RequestFilter


class GzipOption(BaseModel):
    compress_level: int = 6
    compress_types: List[str] = None
    compress_min_length: int = 1024


DEFAULT_COMPRESS_TYPES = [
    'application/json',
    'text/css',
    'text/html',
    'text/javascript',
    'text/xml',
]


class GzipFilter(RequestFilter):

    def __init__(self, **kwargs):
        self.option = GzipOption(**kwargs)
        if not self.option.compress_types:
            self.option.compress_types = DEFAULT_COMPRESS_TYPES

    def after_request(self, response):
        return self._gzip_compress(response)

    def _gzip_compress(self, response):
        if response.mimetype not in self.option.compress_types \
                or response.content_length < self.option.compress_min_length \
                or "Content-Encoding" in response.headers:
            return response

        response.set_data(
            gzip.compress(
                response.get_data(),
                compresslevel=self.option.compress_level,
            )
        )
        response.headers['Content-Encoding'] = 'gzip'
        response.headers['Content-Length'] = response.content_length
        return response



from guniflask.beans.post_processor import BeanPostProcessor
from guniflask.context.bean_context import BeanContext, BeanContextAware


class BeanContextAwareProcessor(BeanPostProcessor):
    def __init__(self, bean_context: BeanContext):
        self._bean_context = bean_context

    def post_process_before_initialization(self, bean, bean_name: str):
        if isinstance(bean, BeanContextAware):
            bean.set_bean_context(self._bean_context)



from abc import ABCMeta, abstractmethod


class Authentication(metaclass=ABCMeta):
    def __init__(self, authorities=None):
        self._authorities = set()
        if authorities is not None:
            self._authorities.update(authorities)
        self._authenticated = False
        self.details = None

    @property
    @abstractmethod
    def name(self):
        pass  # pragma: no cover

    @property
    @abstractmethod
    def principal(self):
        pass  # pragma: no cover

    @property
    @abstractmethod
    def credentials(self):
        pass  # pragma: no cover

    def authenticate(self, value):
        self._authenticated = value

    @property
    def is_authenticated(self):
        return self._authenticated

    @property
    def authorities(self):
        return self._authorities



from functools import update_wrapper

from flask import copy_current_request_context, current_app


def run_with_context(func):
    try:
        func = copy_current_request_context(func)
    except Exception:
        pass
    app_ctx = current_app.app_context()

    def wrapper(*args, **kwargs):
        with app_ctx:
            return func(*args, **kwargs)

    return update_wrapper(wrapper, func)



import logging
from typing import List

from guniflask.beans.errors import BeanCurrentlyInCreationError
from guniflask.beans.lifecycle import DisposableBean

log = logging.getLogger(__name__)


class SingletonBeanRegistry:
    def __init__(self):
        self._singleton_objects = {}
        self._singletons_currently_in_creation = set()
        self._disposable_beans = {}

    def register_singleton(self, bean_name: str, singleton_obj):
        old_object = self._singleton_objects.get(bean_name)
        if old_object is not None:
            raise ValueError(f'Cannot register singleton object [{singleton_obj}], '
                             f'because the bean name "{bean_name}" already exists')
        self._add_singleton(bean_name, singleton_obj)

    def get_singleton(self, bean_name: str):
        return self._singleton_objects.get(bean_name)

    def get_singleton_from_factory(self, bean_name: str, singleton_factory):
        singleton_obj = self._singleton_objects.get(bean_name)
        if singleton_obj is None:
            self._before_singleton_creation(bean_name)
            try:
                singleton_obj = singleton_factory()
            finally:
                self._after_singleton_creation(bean_name)
            if singleton_obj is not None:
                self._add_singleton(bean_name, singleton_obj)
        return singleton_obj

    def contains_singleton(self, bean_name: str) -> bool:
        return bean_name in self._singleton_objects

    def get_singleton_names(self) -> List[str]:
        return list(self._singleton_objects.keys())

    def is_singleton_currently_in_creation(self, bean_name: str) -> bool:
        return bean_name in self._singletons_currently_in_creation

    def register_disposable_bean(self, bean_name: str, bean):
        self._disposable_beans[bean_name] = bean

    def destroy_singletons(self):
        disposable_bean_names = list(self._disposable_beans.keys())
        for bean_name in disposable_bean_names:
            self.destroy_singleton(bean_name)

    def destroy_singleton(self, bean_name: str):
        self._remove_singleton(bean_name)
        if bean_name in self._disposable_beans:
            bean = self._disposable_beans.pop(bean_name)
            self._destroy_bean(bean_name, bean)

    def _before_singleton_creation(self, bean_name):
        if self.is_singleton_currently_in_creation(bean_name):
            raise BeanCurrentlyInCreationError(bean_name)
        self._singletons_currently_in_creation.add(bean_name)

    def _after_singleton_creation(self, bean_name):
        self._singletons_currently_in_creation.remove(bean_name)

    def _add_singleton(self, bean_name, singleton_obj):
        self._singleton_objects[bean_name] = singleton_obj

    def _remove_singleton(self, bean_name):
        if bean_name in self._singleton_objects:
            self._singleton_objects.pop(bean_name)

    def _destroy_bean(self, bean_name, bean: DisposableBean):
        try:
            bean.destroy()
        except Exception:
            log.error(f'Failed to destroy the bean named "{bean_name}"', exc_info=True)



from typing import List, Type


class Annotation:
    def __init__(self, **kwargs):
        self.attributes = {}
        self.attributes.update(kwargs)

    def __contains__(self, key):
        return key in self.attributes

    def __setitem__(self, key, value):
        self.attributes[key] = value

    def __getitem__(self, key):
        return self.attributes.get(key)


class AnnotationMetadata:
    def __init__(self, source):
        if not hasattr(source, '__name__'):
            raise ValueError('Cannot annotate to the object without "__name__" attribute')
        self._source = source
        self._annotations = {}

    @property
    def source(self):
        return self._source

    def add_annotation(self, annotation: Annotation):
        self._annotations[type(annotation)] = annotation

    def get_annotation(self, annotation_type: Type[Annotation]) -> Annotation:
        return self._annotations.get(annotation_type)

    @property
    def annotations(self) -> List[Annotation]:
        return list(self._annotations.values())

    def is_annotated(self, annotation_type: Type[Annotation]) -> bool:
        for a in self._annotations.values():
            if isinstance(a, annotation_type):
                return True
        return False


ANNOTATION_METADATA = '__annotation_metadata'


class AnnotationUtils:
    @staticmethod
    def get_annotation_metadata(source) -> AnnotationMetadata:
        if hasattr(source, ANNOTATION_METADATA):
            metadata = getattr(source, ANNOTATION_METADATA)
            # annotations on class cannot be inherited
            if not isinstance(metadata.source, type) or metadata.source == source:
                return metadata

    @staticmethod
    def get_annotation(source, annotation_type: Type[Annotation]) -> Annotation:
        annotation_metadata = AnnotationUtils.get_annotation_metadata(source)
        if annotation_metadata is not None:
            return annotation_metadata.get_annotation(annotation_type)

    @staticmethod
    def get_annotations(source) -> List[Annotation]:
        annotation_metadata = AnnotationUtils.get_annotation_metadata(source)
        if annotation_metadata is not None:
            return annotation_metadata.annotations
        return []

    @staticmethod
    def add_annotation(source, annotation: Annotation):
        annotation_metadata = AnnotationUtils.get_annotation_metadata(source)
        if annotation_metadata is None:
            annotation_metadata = AnnotationMetadata(source)
            setattr(source, ANNOTATION_METADATA, annotation_metadata)
        annotation_metadata.add_annotation(annotation)



import gevent
from flask import abort

from guniflask.context import service
from guniflask.security import current_user
from guniflask.utils.context import run_with_context
from ..config.jwt_config import jwt_manager


@service
class AccountService:
    accounts = {
        'root': {
            'authorities': ['role_admin'],
            'password': '123456',
        }
    }

    def login(self, username: str, password: str) -> dict:
        if username not in self.accounts or self.accounts[username]['password'] != password:
            return abort(403)
        account = self.accounts[username]
        token = jwt_manager.create_access_token(authorities=account['authorities'], username=username)
        return {
            'username': username,
            'access_token': token,
        }

    def get_account_info(self) -> dict:
        return self._do_get_account_info()

    def get_account_info_by_gevent(self) -> dict:
        job = gevent.spawn(
            run_with_context(
                self._do_get_account_info
            )
        )
        gevent.wait([job])
        return job.get()

    def _do_get_account_info(self) -> dict:
        return {
            'username': current_user.username,
            'authorities': list(current_user.authorities),
        }



from abc import ABCMeta, abstractmethod
from collections import defaultdict
from typing import Type, List, Any, Optional

from guniflask.security_config.security_builder import AbstractSecurityBuilder
from guniflask.security_config.security_configurer import SecurityConfigurer


class ConfiguredSecurityBuilder(AbstractSecurityBuilder, metaclass=ABCMeta):
    def __init__(self):
        AbstractSecurityBuilder.__init__(self)
        self._configurers = defaultdict(list)
        self._shared_objects = {}

    def get_configurers(self, configurer_type: Type[SecurityConfigurer]) -> List[SecurityConfigurer]:
        return self._configurers[configurer_type]

    def get_configurer(self, configurer_type: Type[SecurityConfigurer]) -> Optional[SecurityConfigurer]:
        configurers = self._configurers.get(configurer_type)
        if not configurers:
            return
        if len(configurers) != 1:
            raise ValueError(
                f'Only one configurer excepted for type {configurer_type.__name__}, '
                f'but got {len(configurers)}'
            )
        return configurers[0]

    def apply(self, configurer: SecurityConfigurer):
        configurer.builder = self
        self._add(configurer)
        return configurer

    def set_shared_object(self, shared_type: Type[Any], obj: Any):
        self._shared_objects[shared_type] = obj

    def get_shared_object(self, shared_type: Type[Any]) -> Any:
        return self._shared_objects.get(shared_type)

    def get_shared_objects(self):
        return self._shared_objects

    @abstractmethod
    def _perform_build(self) -> Any:
        pass  # pragma: no cover

    def _do_build(self) -> Any:
        self._before_init()
        self._init()

        self._before_configure()
        self._configure()

        result = self._perform_build()
        return result

    def _before_init(self):
        pass

    def _before_configure(self):
        pass

    def _init(self):
        configurers = self._get_all_configurers()
        for c in configurers:
            c.init(self)

    def _configure(self):
        configurers = self._get_all_configurers()
        for c in configurers:
            c.configure(self)

    def _get_all_configurers(self) -> List[SecurityConfigurer]:
        result = []
        for vlist in self._configurers.values():
            for v in vlist:
                result.append(v)
        return result

    def _add(self, configurer: SecurityConfigurer):
        cls = type(configurer)
        self._configurers[cls].append(configurer)



from typing import List

from guniflask.beans.lifecycle import SmartInitializingSingleton
from guniflask.context.annotation import configuration
from guniflask.security_config.web_security import WebSecurity
from guniflask.security_config.web_security_configurer import WebSecurityConfigurer


@configuration
class WebSecurityConfiguration(SmartInitializingSingleton):
    def __init__(self, configurers: List[WebSecurityConfigurer] = None):
        self.web_security = WebSecurity()
        if configurers:
            for c in configurers:
                self.web_security.apply(c)

    def after_singletons_instantiated(self):
        self.web_security.build()
